/*
								+-----------------------------------+
								|                                   |
								|   ***  Dude mesh generator  ***   |
								|                                   |
								|   Copyright   -tHE SWINe- 2007   |
								|                                   |
								|             Dude.cpp              |
								|                                   |
								+-----------------------------------+
*/


#include <vector>
#include "../berLame_src/CallStack.h"
#ifdef WIN32
#include <windows.h>
#endif
#include <GL/gl.h>
#include <GL/glext.h>
#include <math.h>
#include "../berLame_src/Vector.h"
#include "Dude.h"

static Vector2f v_BezierQuad(Vector2f v_p0, Vector2f v_p1, Vector2f v_p2, Vector2f v_p3, float t)
{
	t = (t > 0)? ((t < 1)? t : 1) : 0;
	return v_p0 * ((1 - t) * (1 - t) * (1 - t)) +
		   v_p1 * (3 * t * (1 - t) * (1 - t)) +
		   v_p2 * (3 * t * t * (1 - t)) +
		   v_p3 * (t * t * t);
}

/*
 *								=== CDudeMesh ===
 */

/*
 *	CDudeMesh::CDudeMesh(float f_base_diameter, int n_subsect_size = 5, int n_side_num = 12)
 *		- default constructor
 *		- creates a new mesh having n_subsect_size vertices in one spline subsection
 *		  (there are three subsections, reasonable value is 5) and n_side_num sides
 *		- f_base_diameter is required base diameter
 *		- call b_Status() to see wheter constructor succeeded
 */
CDudeMesh::CDudeMesh(float f_base_diameter, int n_subsect_size, int n_side_num)
	:p_index(0), f_dude_scale(f_base_diameter / (2 * 1.5f) * .75f)
{
	float f_sh = f_dude_scale * 1.5f * 4; // r to D -> 2, it takes half of shadow tex -> 2 * 2
	v_shadow_size = Vector2f(f_sh, 2 * f_sh);

	int n_section_num = (3 * n_subsect_size + 2);
	n_vertex_num = (n_section_num - 2) * n_side_num + 2;
	n_index_num = n_side_num * (2 + (n_section_num - 2) * 2);

	if((p_vertex = new float[6 * n_vertex_num]) &&
	   (p_index = new unsigned short[n_index_num]) &&
	   !Create(n_section_num, n_subsect_size, n_side_num, f_base_diameter)) {
		delete[] p_vertex;
		p_vertex = 0;
		delete[] p_index;
		p_index = 0;
		// delete to mark error
	}
}

/*
 *	CDudeMesh::~CDudeMesh()
 *		- destructor
 */
CDudeMesh::~CDudeMesh()
{
	if(p_vertex)
		delete[] p_vertex;
	if(p_index)
		delete[] p_index;
}

/*
 *	bool CDudeMesh::b_Status() const
 *		- returns true in case mesh was created successfuly and can be used, otherwise false
 */
bool CDudeMesh::b_Status() const
{
	return p_vertex && p_index;
}

/*
 *	float CDudeMesh::f_Scale() const
 *		- returns mesh scaling factor (radius of dude base)
 */
float CDudeMesh::f_Scale() const
{
	return f_dude_scale;
}

/*
 *	void CDudeMesh::PrepareShadow(Vector2f v_dude_shadow_offset)
 *		- prepares shadow vertex and texture coordinates for shadow texture
 *		  where dude base center is at v_dude_shadow_offset in OpenGL viewport coordinates
 */
void CDudeMesh::PrepareShadow(Vector2f v_dude_shadow_offset)
{
	float f_off_s = (1 + v_dude_shadow_offset.x) * .5f * v_shadow_size.x;
	float f_off_t = (1 + v_dude_shadow_offset.y) * .5f * v_shadow_size.y;
	// calc texture offsets in texturespace

	float f_min_t = (1 + v_dude_shadow_offset.y) * .5f;
	// min t texcoord (so only the part of shadow from base of dude up is drawn)

	p_shadow_vertex[0] = Vector2f(-f_off_s, 0);
	p_shadow_vertex[1] = Vector2f(-f_off_s + v_shadow_size.x, -f_off_t + v_shadow_size.y);
	// calc quad coords (centered arround dude origin in shadow texture)

	p_shadow_texcoord[0] = Vector2f(0, f_min_t);
	p_shadow_texcoord[1] = Vector2f(1, 1);
	// calc minmax texcoords
}

/*
 *	void CDudeMesh::BindVertexArray() const
 *		- binds dude vertex array
 *		- required for drawing
 */
void CDudeMesh::BindVertexArray() const
{
	glInterleavedArrays(GL_N3F_V3F, 6 * sizeof(float), p_vertex);
}

/*
 *	void CDudeMesh::Draw() const
 *		- draws the dude using current matrices
 */
void CDudeMesh::Draw() const
{
	glDrawElements(GL_TRIANGLE_STRIP, n_index_num, GL_UNSIGNED_SHORT, p_index);
}

/*
 *	void CDudeMesh::Draw(Vector2f v_pos) const
 *		- draws dude at position v_pos
 *		- v_org is origin of first stand, v_offset is offset between stands
 */
void CDudeMesh::Draw(Vector2f v_pos) const
{
	glPushMatrix();
	glTranslatef(v_pos.x, 0, v_pos.y);
	glDrawElements(GL_TRIANGLE_STRIP, n_index_num, GL_UNSIGNED_SHORT, p_index);
	glPopMatrix();
}

/*
 *	void CDudeMesh::ReleaseVertexArray() const
 *		- releases vertex array
 *		- call after drawing
 */
void CDudeMesh::ReleaseVertexArray() const
{
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);
}

/*
 *	void CDudeMesh::DrawShadow(Vector2f v_pos, Vector3f v_light_pos) const
 *		- draws dude shadow at position v_pos
 *		- extrude shadow as if cast from light v_light_pos
 *		- note this requires shadow texture to be bound and dude vertex array not to be bound
 */
void CDudeMesh::DrawShadow(Vector2f v_pos, Vector3f v_light_pos) const
{
	Vector3f v_dir(v_pos.x, 0, v_pos.y);
	v_dir -= v_light_pos;
	v_dir.Normalize();
	v_dir /= -v_dir.y / 3; // scale shadow (3 is ad-hoc value)
	v_dir.y = 0;
	// shadow direction

	Vector3f v_x(v_dir.v_Cross(Vector3f(0, -1, 0)));
	v_x.Normalize();
	// shadow right vector

	float f_x00 = v_x.x * p_shadow_vertex[0].x + v_dir.x * p_shadow_vertex[0].y + v_pos.x;
	float f_y00 = v_x.z * p_shadow_vertex[0].x + v_dir.z * p_shadow_vertex[0].y + v_pos.y;
	float f_x10 = v_x.x * p_shadow_vertex[1].x + v_dir.x * p_shadow_vertex[0].y + v_pos.x;
	float f_y10 = v_x.z * p_shadow_vertex[1].x + v_dir.z * p_shadow_vertex[0].y + v_pos.y;
	float f_x11 = v_x.x * p_shadow_vertex[1].x + v_dir.x * p_shadow_vertex[1].y + v_pos.x;
	float f_y11 = v_x.z * p_shadow_vertex[1].x + v_dir.z * p_shadow_vertex[1].y + v_pos.y;
	float f_x01 = v_x.x * p_shadow_vertex[0].x + v_dir.x * p_shadow_vertex[1].y + v_pos.x;
	float f_y01 = v_x.z * p_shadow_vertex[0].x + v_dir.z * p_shadow_vertex[1].y + v_pos.y;
	// calculate corner positions

	glBegin(GL_QUADS);
	glTexCoord2f(p_shadow_texcoord[0].x, p_shadow_texcoord[0].y);
	glVertex3f(f_x00, 0, f_y00);
	glTexCoord2f(p_shadow_texcoord[0].x, p_shadow_texcoord[1].y);
	glVertex3f(f_x01, 0, f_y01);
	glTexCoord2f(p_shadow_texcoord[1].x, p_shadow_texcoord[1].y);
	glVertex3f(f_x11, 0, f_y11);
	glTexCoord2f(p_shadow_texcoord[1].x, p_shadow_texcoord[0].y);
	glVertex3f(f_x10, 0, f_y10);
	glEnd();
	// draw the quad
}

bool CDudeMesh::Create(int n_section_num, int n_subsect_size, int n_side_num, float f_stand_size)
{
	Vector2f *p_list;
	if(!(p_list = new Vector2f[2 * n_section_num]))
		return false;
	Vector2f *p_profile_pos_list = p_list;
	Vector2f *p_profile_norm_list = p_list + n_section_num;
	// alloc lists for dude profile normals and positions

	for(int i = 0; i <= n_subsect_size; ++ i) {
		float t = i * (1.0f / n_subsect_size);
		const float k = 0.5522847498f;
		if(i < n_subsect_size) {
			p_profile_pos_list[i] = v_BezierQuad(Vector2f(0, 7), Vector2f(k, 7),
				Vector2f(1, 6 + k), Vector2f(1, 6), t) * f_dude_scale;
			p_profile_pos_list[n_subsect_size + i] = v_BezierQuad(Vector2f(1, 6), Vector2f(1, 6 - k),
				Vector2f(.25f, 5), Vector2f(1, 4), t) * f_dude_scale;
		}
		p_profile_pos_list[n_subsect_size * 2 + i] = v_BezierQuad(Vector2f(1, 4), Vector2f(1.75f, 3),
			Vector2f(1.5f, 2), Vector2f(1.5f, 0), t) * f_dude_scale;
	}
	p_profile_norm_list[0] = Vector2f(0, 1);
	p_profile_pos_list[n_section_num - 1] = Vector2f(0, 0);
	p_profile_norm_list[n_section_num - 1] = Vector2f(0, -1);
	for(int i = 1; i < n_section_num - 1; ++ i) {
		p_profile_norm_list[i] = -(p_profile_pos_list[i - 1] - p_profile_pos_list[i]).v_Orthogonal();
		p_profile_norm_list[i].Normalize();
	}
	// generate dude profile, along with normals

	float *p_dest_vert = p_vertex;
	*p_dest_vert = 0;
	*(++ p_dest_vert) = -1;
	*(++ p_dest_vert) = 0;
	*(++ p_dest_vert) = 0;
	*(++ p_dest_vert) = 0;
	*(++ p_dest_vert) = 0; // bottom
	*(++ p_dest_vert) = 0;
	*(++ p_dest_vert) = 1;
	*(++ p_dest_vert) = 0;
	*(++ p_dest_vert) = 0;
	*(++ p_dest_vert) = 7 * f_dude_scale;
	*(++ p_dest_vert) = 0; // top
	for(int i = 0; i < n_side_num; ++ i) {
		float f_angle = (6.28f / n_side_num) * i;
		float f_cos = (float)cos(f_angle);
		float f_sin = (float)sin(f_angle);

		for(int j = 1; j < n_section_num - 1; ++ j) {
			Vector3f v_norm(f_cos * p_profile_norm_list[j].x,
				p_profile_norm_list[j].y, f_sin * p_profile_norm_list[j].x);
			v_norm.Normalize();

			p_dest_vert[1] = v_norm.x;
			p_dest_vert[2] = v_norm.y;
			p_dest_vert[3] = v_norm.z;
			p_dest_vert[4] = f_cos * p_profile_pos_list[j].x;
			p_dest_vert[5] = p_profile_pos_list[j].y;
			p_dest_vert[6] = f_sin * p_profile_pos_list[j].x;
			p_dest_vert += 6;
		}
	}
	_ASSERTE(p_dest_vert == p_vertex + 6 * n_vertex_num);
	// create dude vertex array (GL_N3F_V3F)

	unsigned short *p_dest_idx = p_index;
	for(int i = 0; i < n_side_num; ++ i) {
		int n_base0 = 2 + i * (n_section_num - 2);
		int n_base1 = 2 + ((i + 1) % n_side_num) * (n_section_num - 2);
		// base vertex indices for two rotated profiles

		if(i & 1) {
			*p_dest_idx ++ = 1;
			for(int j = 0; j < (n_section_num - 2) * 2; ++ j)
				*p_dest_idx ++ = ((j & 1)? n_base1 : n_base0) + j / 2;
			*p_dest_idx ++ = 0;
		} else {
			*p_dest_idx ++ = 0;
			for(int j = (n_section_num - 2) * 2 - 1; j >= 0; -- j)
				*p_dest_idx ++ = ((j & 1)? n_base1 : n_base0) + j / 2;
			*p_dest_idx ++ = 1;
		}
	}
	_ASSERTE(p_dest_idx == p_index + n_index_num);
	// generate vertex indices

	delete[] p_list;
	// free temp vertex lists

	return true;
}

/*
 *								=== ~CDudeMesh ===
 */
