/*
								+----------------------------------+
								|                                  |
								| ***  Procedural texture gen  *** |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|             ProcTex.h            |
								|                                  |
								+----------------------------------+
*/

#ifndef __PROCEDURAL_TEXTURE_GENERATOR_INCLUDED
#define __PROCEDURAL_TEXTURE_GENERATOR_INCLUDED

/*
 *	class CProcTexGen
 *		- class with static procedural texture generation functions
 */
class CProcTexGen {
public:
	/*
	 *	static CGLTexture_2D *CProcTexGen::p_BuildTexture2D(CGLState *p_state, int n_width,
	 *		int n_height, GLenum n_internal_format, Vector4f (*p_tex_func)(Vector2f v_pos))
	 *		- builds procedural texture n_width per n_height pixels of format n_internal_format
	 *		- pixel values are generated by p_tex_func function which is passed relative texcoords
	 *		  (in range [0, 1] or more precisely [1 / pixel-size, 1 - 1 / pixel-size])
	 *		- note p_tex_func can be 0 to create uninitialized texture
	 *		- p_state is OpenGL state guard
	 *		- returns a new texture object on success, 0 on failure
	 */
	static CGLTexture_2D *p_BuildTexture2D(CGLState *p_state, int n_width, int n_height,
		GLenum n_internal_format, Vector4f (*p_tex_func)(Vector2f v_pos));

	/*
	 *	static inline void CProcTexGen::WebColor(unsigned int n_value, Vector4f &r_v_color)
	 *		- converts from hexadecimal RGB color n_value to RGBA vector r_v_color (A = 1)
	 *		- this is just artist convenience tool to ease photoshop palette copy-pasting
	 */
	static inline void WebColor(unsigned int n_value, Vector4f &r_v_color)
	{
		int	n_Red = (n_value >> 16) & 0xff,
			n_Green = (n_value >> 8) & 0xff,
			n_Blue = n_value & 0xff;
		const float f_scale = 1.0f / 255;
		r_v_color.x = n_Red * f_scale;
		r_v_color.y = n_Green * f_scale;
		r_v_color.z = n_Blue * f_scale;
		r_v_color.w = 1;
	}

	/*
	 *	static float CProcTexGen::f_Turbulence(Vector2f v_pos, float f_scale,
	 *		int n_octave_num, float f_durability)
	 *		- returns turbulence noise at 2D position v_pos with scale f_scale
	 *		  (spatial multiplier), n_octave_num octaves and f_durability durability
	 *		- note noise is normalized so it gives the same output magnitude at different
	 *		  durabilities and octave numbers
	 */
	static float f_Turbulence(Vector2f v_pos, float f_scale,
		int n_octave_num, float f_durability);

	/*
	 *	static Vector3f CProcTexGen::v_Turbulence(Vector2f v_pos, float f_scale,
	 *		int n_octave_num, float f_durability)
	 *		- returns turbulence noise vector (three different unrelated noise functions)
	 *		  at 2D position v_pos with scale f_scale (spatial multiplier), n_octave_num
	 *		  octaves and f_durability durability
	 *		- note noise is normalized so it gives the same output magnitude at different
	 *		  durabilities and octave numbers
	 */
	static Vector3f v_Turbulence(Vector2f v_pos, float f_scale,
		int n_octave_num, float f_durability);

	/*
	 *	static Vector4f CProcTexGen::v_BoardPaperFunc(Vector2f v_pos)
	 *		- this can be used as texture generation function
	 *		- creates paper-like texture with thin red frame
	 */
	static Vector4f v_BoardPaperFunc(Vector2f v_pos);

	/*
	 *	static Vector4f CProcTexGen::v_StandPaperFunc(Vector2f v_pos, Vector4f v_paper_color)
	 *		- this can be used as texture generation function, except for v_paper_color
	 *		  parameter (requires adaptor function)
	 *		- creates paper-like texture with thin red circle and v_paper_color dot inside the
	 *		  circle
	 *		- note for correct antialiassing this requires texture to have size 32 per 23 pixels
	 */
	static Vector4f v_StandPaperFunc(Vector2f v_pos, Vector4f v_paper_color);

	/*
	 *	static Vector4f CProcTexGen::v_WoodPlankFunc(Vector2f v_pos)
	 *		- this can be used as texture generation function
	 *		- creates wood plank texture (varnished pine wood taken as the paradigm)
	 */
	static Vector4f v_WoodPlankFunc(Vector2f v_pos);
};

#endif //__PROCEDURAL_TEXTURE_GENERATOR_INCLUDED
